/**
  *  \file client/dlg-tech.cc
  *  \brief Tech Upgrade dialog
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/dlg-tech.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/global.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/util.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/misc.h"
#include "game/undo.h"
#include "game/actions/basetech.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/numbersel.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "game/planettransfer.h"
#include "ui/widgets/quit.h"

namespace {
    // bar for selection of a tech level
    class WTechBar : public UINumberSelector {
        string_t name;

        GfxRect  getBarPosition() const;
     public:
        WTechBar(int id, int low, int high, string_t name);
        ~WTechBar();

        void drawContent(GfxCanvas& can);
        bool handleEvent(const UIEvent& e, bool second_pass);
        void getLayoutInfo(LayoutInfo& info);
        void onStateChange(int state, bool enabled);
    };

    // tech upgrade dialog window
    class WTechUpgradeWindow : public UIWindow {
        WCostDisplay*      cost_dpy;
        UINumberSelector*  tech[4];
        GStarbaseTechUpgradeAction& action;
        GStarbaseBuildTransaction& ta;

        void updateTechBar(int);
        void maybeStop(int);
        WTechUpgradeWindow(GStarbaseTechUpgradeAction& action,
                           GStarbaseBuildTransaction& ta);
     public:
        static void doWindow(GPlanet& planet);
    };
}

/******************************** WTechBar *******************************/

/** \class WTechBar
    \brief Tech level bar and Label

    This bar displays a tech level. It is based on UINumberSelector to
    inherit the behaviour; this means it cannot use
    GStarbaseTechUpgradeAction::isValidTech. Actually, this does not use
    GStarbaseTechUpgradeAction at all.

    This widget also has a focus frame. */

/** Constructor.
    \param id,low,high  as for UINumberSelector
    \param name         name of this tech area */
WTechBar::WTechBar(int id, int low, int high, string_t name)
    : UINumberSelector(id, low, high, 1), name(name)
{ }

WTechBar::~WTechBar()
{ }

/** Get position of the actual bar. */
GfxRect
WTechBar::getBarPosition() const
{
    GfxRect r = getExtent();
    r.grow(-2, -2);

    int reduce = r.w / 2;
    r.x += reduce;
    r.w -= reduce;
    return r;
}

void
WTechBar::drawContent(GfxCanvas& can)
{
    static const int dif[7] = { -9, -6, -3, 0, +4, +7, +10 };

    GfxRect r = getBarPosition(), x = getExtent();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    /* frame */
    ctx.setColor(getFocusState() != NoFocus ? COLOR_BLACK : COLOR_GRAY);
    drawRectangle(ctx, x);

    /* text */
    const int wi = 2 * font_ems[FONT_NORMAL];
    drawSolidBar(ctx, GfxRect(r.x - 5 - wi, x.y + 1, wi, x.h - 2), COLOR_GRAY);
    ctx.setColor(COLOR_BLACK);
    ctx.setTextAlign(0, 1);
    ctx.useFont(*fonts[FONT_NORMAL]);
    outTextF(ctx,
             x.x + 3, x.y + x.h/2, r.x - x.x,  // not a hard limit
             name);
    ctx.setTextAlign(2, 1);
    outTextF(ctx,
             r.x - 5, x.y + x.h/2, wi,
             numToString(getValue()));

    /* the tech bar */
    drawFrameDown(can, r);
    r.grow(-1, -1);

    int xt = r.w * getMin() / 10;
    int xs = r.w * getValue() / 10;

    if (getMin() == getValue()) {
        drawSolidBar(ctx, GfxRect(r.x, r.y, xt, r.h), COLOR_GREENBLACK);
    } else {
        drawSolidBar(ctx, GfxRect(r.x, r.y, xt-9, r.h), COLOR_GREENBLACK);
        for (int i = 0; i < 6; ++i) {
            drawSolidBar(ctx, GfxRect(r.x + xt + dif[i], r.y,
                                      dif[i+1]-dif[i],   r.h),
                         COLOR_GREENSCALE6 + i);
        }
        drawSolidBar(ctx, GfxRect(r.x + xt + 10, r.y,
                                  xs - xt - 10,  r.h),
                     COLOR_GREEN);
    }
    if (xs < r.w)
        drawSolidBar(ctx, GfxRect(r.x + xs, r.y, r.w - xs, r.h), COLOR_BLACK);

    int xlim = r.w * getMax() / 10;
    if (xlim < r.w) {
        can.drawBar(GfxRect(r.x + xlim, r.y, r.w - xlim, r.h),
                    standard_colors[COLOR_GRAY],
                    GFX_TRANSPARENT,
                    LTSLASH_FILL,
                    GFX_ALPHA_OPAQUE);
    }

    can.update(getExtent());
}

/** Event handler. Handles clicks into the bar, and numeric input. */
bool
WTechBar::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && hasState(st_Focused)) {
        if (e.key.code >= '1' && e.key.code <= '9') {
            setSelectedWidget();
            setValue(e.key.code - '0');
            return true;
        } else if (e.key.code == '0') {
            setSelectedWidget();
            setValue(10);
            return true;
        }
    } else if (e.type == ev_Mouse && e.mouse.b && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        setSelectedWidget();
        setFocusedWidget();
        GfxRect r = getBarPosition();
        if (r.contains(e.mouse.x, e.mouse.y))
            setValue(1 + (e.mouse.x - r.x) * 10 / r.w);
        return true;
    }
    return UINumberSelector::handleEvent(e, second_pass);
}

void
WTechBar::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.pref_size = info.min_size = GfxPoint(2*20*10, 20);
}

void
WTechBar::onStateChange(int state, bool enabled)
{
    if (state & st_Focused)
        drawWidget(false);
    UINumberSelector::onStateChange(state, enabled);
}

/*************************** WTechUpgradeWindow **************************/

/** \class WTechUpgradeWindow
    \brief Tech Upgrade Window

    This is mainly a container for signal handlers, and initialisation. */

WTechUpgradeWindow::WTechUpgradeWindow(GStarbaseTechUpgradeAction& action,
                                       GStarbaseBuildTransaction& ta)
    : UIWindow(_("Tech Upgrade"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      action(action), ta(ta)
{
    setLayout(UIVBoxLayout::instance5);
}

/** Do tech upgrade dialog for a planet. Creates the window, runs it,
    and does the upgrade.
    \see doTechUpgradeDialog */
void
WTechUpgradeWindow::doWindow(GPlanet& planet)
{
    static const char*const tech_names[] = {
        N_("Engines:"),
        N_("Hulls:"),
        N_("Beam Weapons:"),
        N_("Torpedoes:")
    };

    // FIXME: use GMultiTransaction
    GPlanetTransfer             transfer(planet);
    GStarbaseTechUpgradeAction  action(planet);
    GStarbaseBuildTransaction   ta(action, transfer);

    {
        // FIXME: to ensure proper concurrency support, we'd have to
        // re-evaluate this whenever anything is built at the planet's X,Y's
        // FIXME: references getDisplayedTurn()
        GUndoInformation undo;
        undo.computeInformation(getDisplayedTurn(), planet.getId());
        action.setUndoInformation(undo);
    }

    ta.update();

    ObjectHolder h;

    WTechUpgradeWindow win(action, ta);
    WCostDisplay cost_dpy(ta, GCargoTypeSet(el_Money));
    win.setLayout(UIVBoxLayout::instance5);
    win.add(h.add(new UIFocusIterator(100, 103, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Tab)));

    for (int i = 0; i < 4; ++i) {
        /* if the current value is somewhere outside what we believe to be the valid range,
           fix the valid range. This happens, for example, in unregistered games when a
           tech level is 10 although the rule maximum is 6. */
        WTechBar& bar = h.add(new WTechBar(100 + i,
                                           std::min(action.getTech(i), action.getMinTech(i)),
                                           std::max(action.getTech(i), action.getMaxTech(i)),
                                           _(tech_names[i])));
        bar.setValue(action.getTech(i));
        bar.sig_changed.add(&win, &WTechUpgradeWindow::updateTechBar);
        win.add(bar);
        win.tech[i] = &bar;
    }
    win.add(cost_dpy);
    win.cost_dpy = &cost_dpy;

    UIButton ok_button(cm_Enter,  _("OK"),     SDLK_RETURN);
    ok_button.sig_fired.add(&win, &WTechUpgradeWindow::maybeStop);

    UIGroup hg(UIHBoxLayout::instance5);
    hg.add(h.add(new UISpacer()));
    hg.add(ok_button);
    hg.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    win.add(hg);
    win.add(h.add(new UIQuit(cm_Escape)));
    win.pack();

    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);
    if (rv == cm_Enter) {
        ta.doCommit();
        transfer.doCommit();
    } else {
        ta.doCancel();
        transfer.doCancel();
    }
}

/** Update a bar. Called when a tech bar moves. This updates the
    underlying action and redraws the cost display.
    \param id identifier of tech bar (100 .. 103) */
void
WTechUpgradeWindow::updateTechBar(int id)
{
    id -= 100;
    ASSERT(id >= 0 && id < 4);
    ASSERT(action.isValidTech(id, tech[id]->getValue()));
    action.setTech(id, tech[id]->getValue());
    cost_dpy->drawWidget(false);
}

/** Stop dialog, if permitted. This is the handler for the "OK"
    button. */
void
WTechUpgradeWindow::maybeStop(int /*id*/)
{
    if (!action.getError() && !ta.getError())
        stop(cm_Enter);
}

/****************************** WCostDisplay *****************************/

/** \class WCostDisplay
    \brief Cost Display

    This is a generic display for costs of a transaction (see
    GCargoCostTransaction). It can display any subset of the costs
    (i.e. "tritanium only"). This is intended to be used for "smaller"
    transactions with "simple" costs, like structures or tech levels.
    When buying complicated things like ships or ammo, use
    WItemCostDisplay.

    As an exception, you can't display money without supplies, because
    of automatic supply sale. */

/** Constructor.
    \param action underlying transaction.
    \param mask bitfield of (1 << el_Foo) of what to display. */
WCostDisplay::WCostDisplay(GCargoCostTransaction& action, GCargoTypeSet mask)
    : UIWidget(0), action(action), mask(mask)
{
    if (this->mask.contains(el_Money))
        this->mask += el_Supplies;
}

/** Count lines in display. */
int
WCostDisplay::getLines() const
{
    return mask.contains(el_Tritanium)
        + mask.contains(el_Duranium)
        + mask.contains(el_Molybdenum)
        + mask.contains(el_Supplies)
        + mask.contains(el_Money);
}

/** Display. This roughly equals the PCC 1.x functions
    InitBill, ShowCost, ShowCash. */
void
WCostDisplay::drawContent(GfxCanvas& can)
{
    static const GCargoType elements[] = {
        el_Tritanium, el_Duranium, el_Molybdenum, el_Money, el_Supplies
    };
    enum { Index_Money = 3 };

    GfxRect x = getExtent();
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());

    // FIXME: currently, we're using a fixed layout as in 1.x:
    // InitBill(105, 270, 535) => 430 wide
    //   You need___________  You have_____________________
    //   ^0             150^  ^165                     430^
    // ShowCash(105, 290, ...)
    //   Cash           NNNN  XXXX     NNN remaining
    //   Supplies       NNNN  XXXX
    //   ^0             150^  215^     ^230

    const int m  = x.x + 165;
    const int x3 = x.x + 230;
    const int w3 = x.w - 230;

    /* Headings */
    ctx.setColor(UIColor::tc_Static);
    ctx.useFont(*fonts[FONT_NORMAL]);
    outText(ctx,   x.x, x.y, _("You need:"));
    outText(ctx,   m,   x.y, _("You have:"));
    drawHLine(ctx, x.x, x.y + font_heights[FONT_NORMAL], m - 15);
    drawHLine(ctx, m,   x.y + font_heights[FONT_NORMAL], x.x + x.w);

    /* Minerals */
    int y1 = extent.y + font_heights[FONT_NORMAL] * 3/2;
    for (unsigned i = 0; i < countof(elements); ++i) {
        const GCargoType ele = elements[i];

        if (!mask.contains(ele))
            continue;

        /* left side */
        ctx.setSolidBackground();
        ctx.setColor(UIColor::tc_Static);
        ctx.setTextAlign(0, 0);
        outTextF(ctx, x.x, y1, 150, getCargoName(ele) + ":");
        ctx.setTextAlign(2, 0);
        outTextF(ctx, m+50, y1, 50, numToString(action.getRemainingCargo(ele) + action.getReservedCargo(ele)));
        ctx.setTransparentBackground();  // Background erasure is done by first outTextF
        outTextF(ctx, x.x+150, y1, 150, numToString(action.getCost(ele)));
        ctx.setTextAlign(0, 0);

        /* right side */
        ctx.setSolidBackground();
        if (i < Index_Money || action.getCost(el_Supplies) != 0) {
            /* it's a mineral, or we're using supplies for this buy.
               In the latter case, we do per-resource accounting, too. */
            if (action.getMissingCargo(ele)) {
                ctx.setColor(UIColor::tc_Red);
                outTextF(ctx, x3, y1, w3, format(_("%d kt too little"), numToString(action.getMissingCargo(ele))));
            } else {
                ctx.setColor(UIColor::tc_Green);
                outTextF(ctx, x3, y1, w3, format(_("%d kt remaining"), numToString(action.getRemainingCargo(ele))));
            }
        } else if (i == Index_Money) {
            ASSERT(ele == el_Money);
            if (action.getMissingCargo(el_Money)) {
                ctx.setColor(UIColor::tc_Red);
                outTextF(ctx, x3, y1, w3,
                         /* translators: this is a two-line message, the first
                            line must contain the %d placeholder */
                         format(_("You need %d mc"), numToString(action.getMissingCargo(el_Money))));
                outTextF(ctx, x3, y1 + font_heights[FONT_NORMAL], w3,
                         _("and/or supplies more."));
            } else {
                ctx.setColor(UIColor::tc_Green);
                outTextF(ctx, x3, y1, w3,
                         format(_("%d mc remaining"), numToString(action.getRemainingCargo(el_Money))));
                outTextF(ctx, x3, y1 + font_heights[FONT_NORMAL], w3,
                         format(_("%d kt remaining"), numToString(action.getRemainingCargo(el_Supplies))));
            }
        } else {
            /* when we're here, it's the supplies line but the item
               we're buying doesn't cost supplies, and everything
               needed has already been displayed. */
            ASSERT(ele == el_Supplies && action.getCost(el_Supplies) == 0);
        }
        y1 += font_heights[FONT_NORMAL];
    }

    can.update(getExtent());
}

void
WCostDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size =
        GfxPoint(430, (2*getLines() + 3) * font_heights[FONT_NORMAL] / 2);
}

/** Tech upgrade dialog. Does the dialog and all necessary updates. */
void
doTechUpgradeDialog(GPlanet& planet)
{
    WTechUpgradeWindow::doWindow(planet);
}
