/**
  *  \file client/search.cc
  *  \brief Search Function Main Code
  *
  *  This module implements the back-end of the search function:
  *  storage, parsing and execution of search queries.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/search.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "int/bytecode.h"
#include "int/exec.h"
#include "int/expr.h"
#include "int/if/ionif.h"
#include "int/if/mineif.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "int/if/ufoif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "int/lex.h"
#include "int/parseexpr.h"
#include "int/process.h"
#include "int/unary.h"
#include "ui/msgbox.h"
#include "util/misc.h"
#include "util/nls.h"
#include "game/team.h"


/** Prepare a search query. Parses a TUserSearchQuery into a TParsedSearchQuery.
    On error, displays a dialog box, and returns false. On success, returns true.
    \param query [out] Parsed query
    \param in    [in] User query */
bool
prepareSearchQuery(TParsedSearchQuery& query, const TUserSearchQuery& in)
{
    /* Prepare object types */
    query.search_obj = in.search_obj;

    /* Prepare query string */
    string_t expr = strTrim(in.query);
    if (expr.size() == 0) {
        /* Make a "match any" query */
        query.match = query.MatchAny;
        return true;
    }

    switch (in.search_kind) {
     case TUserSearchQuery::MatchName:
        query.name = strUCase(expr);
        if (!stringValue(expr, query.id))
            if (!(expr[0] == '#' && stringValue(expr.substr(1), query.id)))
                query.id = -1;
        query.match = query.MatchName;
        return true;

     case TUserSearchQuery::MatchTrue:
     case TUserSearchQuery::MatchFalse:
        try {
            // Parse expression
            IntTokenizer tok(expr);
            std::auto_ptr<IntExprNode> node(IntExpressionParser(tok).parse());
            if (tok.getCurrentToken() != tok.tEnd)
                throw IntError::garbageAtEnd(true);

            // Compile expression
            query.expr_bco = new IntBytecodeObject();
            node->compileValue(*query.expr_bco, IntCompilationContext());

            // Negate if necessary
            if (in.search_kind == TUserSearchQuery::MatchFalse)
                query.expr_bco->addInstruction(IntOpcode::maUnary, unNot2, 0);
            query.expr_bco->relocate();

            // FIXME: this sets the BCO to "no result", although it produces one.
            // Since it will be called with a "no result" frame, it'll work by
            // accident, but it will fail when the interpreter enforces stack
            // cleanliness.

            query.match = query.MatchExpr;
        }
        catch (IntError& e) {
            messageBox(e.what(), _("Search Object"));
            return false;
        }
        return true;

     case TUserSearchQuery::MatchLocation:
        if (!query.location.parseCoordinates(expr)) {
            messageBox(_("Please enter a valid X,Y coordinate pair."), _("Search Object"));
            return false;
        }
        query.match = query.MatchLocation;
        return true;

     default:
        return false;
    }
}

/*************************** executeSearchQuery **************************/

static bool
evaluateProcess(IntExecutionContext& exec, TSearchResultStatus& status)
{
    if (runTemporaryProcess(exec)) {
        status.did_evaluate = true;
        return getBoolValue(exec.getResult()) > 0;
    } else if (exec.getState() == IntExecutionContext::psError) {
        status.had_error = true;
        status.last_error = exec.getError().what();
        return false;
    } else {
        return false;
    }
}

/** Search planets.
    \param query  [in] Search query
    \param result [out] Append result here
    \param status [out] Update status here
    \param mustHaveBase [in] true to restrict search on starbases only */
static void
searchPlanets(const TParsedSearchQuery& query, GObjectList& result, TSearchResultStatus& status, bool mustHaveBase)
{
    for (int pid = current_anyplanet_type.findNextIndex(0); pid != 0; pid = current_anyplanet_type.findNextIndex(pid)) {
        bool match;
        GPlanet* pp = dynamic_cast<GPlanet*>(&current_anyplanet_type.getObjectByIndex(pid));
        if (!pp)
            continue;
        GPlanet& p = *pp;
        if (mustHaveBase && !p.hasBase())
            continue;
        if ((query.search_obj & TUserSearchQuery::SearchMineOnly) != 0 && !p.isPlayable(GPlanet::Playable))
            continue;

        status.had_object = true;
        switch (query.match) {
         case TParsedSearchQuery::MatchAny:
            match = p.hasAnyPlanetData();
            break;
         case TParsedSearchQuery::MatchName:
            match = pid == query.id
                || strUCase(p.getName(GObject::PlainName)).find(query.name) != string_t::npos
                || strUCase(IntValue::toStringOf(getPlanetProperties(pid)[pp_Comment], false)).find(query.name) != string_t::npos;
            break;
         case TParsedSearchQuery::MatchLocation:
            match = p.getPos() == query.location;
            break;
         case TParsedSearchQuery::MatchExpr:
            {
                IntExecutionContext exec("Temporary: Search");
                exec.pushNewContext(new IntPlanetContext(pid));
                exec.pushNewFrame(new IntExecutionFrame(query.expr_bco));
                match = evaluateProcess(exec, status);
            }
            break;
         default:
            match = false;
            break;
        }

        if (match) {
            result.addObject(current_anyplanet_type, pid);
        }
    }
}

/** Search starships.
    \param query  [in] Search query
    \param result [out] Append result here
    \param status [out] Update status here */
static void
searchShips(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result)
{
    for (int sid = current_anyship_type.findNextIndex(0); sid != 0; sid = current_anyship_type.findNextIndex(sid)) {
        GShip* ps = dynamic_cast<GShip*>(&current_anyship_type.getObjectByIndex(sid));
        if (!ps)
            continue;
        GShip& s = *ps;
        if ((query.search_obj & TUserSearchQuery::SearchMineOnly) != 0 && !s.isPlayable(GShip::Playable))
            continue;
        bool match;
        status.had_object = true;
        switch (query.match) {
         case TParsedSearchQuery::MatchAny:
            match = true;
            break;
         case TParsedSearchQuery::MatchName:
            match = sid == query.id
                || strUCase(s.getName(GObject::PlainName)).find(query.name) != string_t::npos
                || strUCase(IntValue::toStringOf(getShipProperties(sid)[sp_Comment], false)).find(query.name) != string_t::npos;
            break;
         case TParsedSearchQuery::MatchLocation:
            match = s.getPos() == query.location;
            break;
         case TParsedSearchQuery::MatchExpr:
            {
                IntExecutionContext exec("Temporary: Search");
                exec.pushNewContext(new IntShipContext(sid));
                exec.pushNewFrame(new IntExecutionFrame(query.expr_bco));
                match = evaluateProcess(exec, status);
            }
            break;
         default:
            match = false;
            break;
        }

        if (match) {
            result.addObject(current_anyship_type, sid);
        }
    }
}

/** Search Ufos.
    \param query  [in] Search query
    \param result [out] Append result here
    \param status [out] Update status here */
static void
searchUfos(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result)
{
    if ((query.search_obj & TUserSearchQuery::SearchMineOnly) != 0)
        return;
    GUfoType& con = getDisplayedTurn().getCurrentUniverse().ty_ufos;
    for (GUfoType::iterator it = con.begin(); it != con.end(); ++it) {
        bool match;
        status.had_object = true;
        switch (query.match) {
         case TParsedSearchQuery::MatchAny:
            match = true;
            break;
         case TParsedSearchQuery::MatchName:
            match = it->getId() == query.id || strUCase(it->getName(GObject::PlainName)).find(query.name) != string_t::npos;
            break;
         case TParsedSearchQuery::MatchLocation:
            match = distanceSquared(it->getPos(), query.location) <= square(it->getRadius());
            break;
         case TParsedSearchQuery::MatchExpr:
            {
                IntExecutionContext exec("Temporary: Search");
                exec.pushNewContext(new IntUfoContext(con, it));
                exec.pushNewFrame(new IntExecutionFrame(query.expr_bco));
                match = evaluateProcess(exec, status);
            }
            break;
         default:
            match = false;
            break;
        }

        if (match) {
            // FIXME: this sucks.
            result.addObject(current_ufo_type, 1 + std::distance(con.begin(), it));
        }
    }
}

/** Search minefields.
    \param query  [in] Search query
    \param result [out] Append result here
    \param status [out] Update status here */
static void
searchMinefields(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result)
{
    for (int mid = current_minefield_type.findNextIndex(0); mid != 0; mid = current_minefield_type.findNextIndex(mid)) {
        bool match;
        if ((query.search_obj & TUserSearchQuery::SearchMineOnly) != 0)
            if (GMinefield* mf = dynamic_cast<GMinefield*>(&current_minefield_type.getObjectByIndex(mid)))
                if (mf->getOwner() != getPlayerId())
                    continue;
        status.had_object = true;
        switch (query.match) {
         case TParsedSearchQuery::MatchAny:
            match = true;
            break;
         case TParsedSearchQuery::MatchName:
            match = mid == query.id;
            break;
         case TParsedSearchQuery::MatchLocation:
            if (GMinefield* mf = dynamic_cast<GMinefield*>(&current_minefield_type.getObjectByIndex(mid))) {
                match = distanceSquared(mf->getPos(), query.location) <= mf->getUnits();
            } else {
                match = false;
            }
            break;
         case TParsedSearchQuery::MatchExpr:
            {
                IntExecutionContext exec("Temporary: Search");
                exec.pushNewContext(new IntMinefieldContext(mid));
                exec.pushNewFrame(new IntExecutionFrame(query.expr_bco));
                match = evaluateProcess(exec, status);
            }
            break;
         default:
            match = false;
            break;
        }

        if (match) {
            result.addObject(current_minefield_type, mid);
        }
    }
}

/** Search ion storms.
    \param query  [in] Search query
    \param result [out] Append result here
    \param status [out] Update status here */
static void
searchIonStorms(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result)
{
    if ((query.search_obj & TUserSearchQuery::SearchMineOnly) != 0)
        return;
    for (int iid = current_ion_type.findNextIndex(0); iid != 0; iid = current_ion_type.findNextIndex(iid)) {
        GIonStorm* pi = dynamic_cast<GIonStorm*>(&current_ion_type.getObjectByIndex(iid));
        if (!pi) {
            continue;
        }
        GIonStorm& i = *pi;
        bool match;
        status.had_object = true;
        switch (query.match) {
         case TParsedSearchQuery::MatchAny:
            match = true;
            break;
         case TParsedSearchQuery::MatchName:
            match = (iid == query.id) || strUCase(i.getName(GObject::PlainName)).find(query.name) != string_t::npos;
            break;
         case TParsedSearchQuery::MatchLocation:
            match = distanceSquared(i.getPos(), query.location) <= square(i.getRadius());
            break;
         case TParsedSearchQuery::MatchExpr:
            {
                IntExecutionContext exec("Temporary: Search");
                exec.pushNewContext(new IntIonContext(iid));
                exec.pushNewFrame(new IntExecutionFrame(query.expr_bco));
                match = evaluateProcess(exec, status);
            }
            break;
         default:
            match = false;
            break;
        }

        if (match) {
            result.addObject(current_ion_type, iid);
        }
    }
}

/** Execute search query.
    \param query  [in] Parsed query
    \param status [out] Result status
    \param result [out] Result */
void
executeSearchQuery(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result)
{
    // Search planets and bases
    if (query.search_obj & TUserSearchQuery::SearchPlanets) {
        searchPlanets(query, result, status, false);
    } else if (query.search_obj & TUserSearchQuery::SearchBases) {
        searchPlanets(query, result, status, true);
    } else {
        // Don't search planets
    }

    // Search ships
    if (query.search_obj & TUserSearchQuery::SearchShips) {
        searchShips(query, status, result);
    }

    // Search Ufos
    if (query.search_obj & TUserSearchQuery::SearchUfos) {
        searchUfos(query, status, result);
    }

    // Search rest
    if (query.search_obj & TUserSearchQuery::SearchOthers) {
        searchMinefields(query, status, result);
        searchIonStorms(query, status, result);
    }
}
