/**
  *  \file client/widgets/optiongrid.cc
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <ctype.h>
#include "client/widgets/optiongrid.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "gfx/font.h"

class WOptionGrid::Item {
 public:
    Item(WOptionGrid& parent, int id, char key, string_t label)
        : parent(parent),
          button(string_t(1, key), tolower(key)),
          id(id),
          key(tolower(key)),
          enabled(true),
          label(label),
          value(),
          font(FONT_NORMAL)
        { }

    WOptionGrid& parent;
    UIButton button;
    int id;
    char key;
    bool enabled;
    string_t label;
    string_t value;
    int font;
};

/** Change item value's font.
    \param font Font number */
WOptionGrid::Ref&
WOptionGrid::Ref::setFont(int font)
{
    if (pItem && pItem->font != font) {
        pItem->font = font;
        pItem->parent.drawWidget();
    }
    return *this;
}

/** Change item value. */
WOptionGrid::Ref&
WOptionGrid::Ref::setValue(string_t str)
{
    if (pItem && pItem->value != str) {
        pItem->value = str;
        pItem->parent.drawWidget();
    }
    return *this;
}

/** Change item value. */
WOptionGrid::Ref&
WOptionGrid::Ref::setValue(const char* str)
{
    return setValue(string_t(str));
}

/** Change item label. */
WOptionGrid::Ref&
WOptionGrid::Ref::setLabel(string_t str)
{
    if (pItem && pItem->label != str) {
        pItem->label = str;
        pItem->parent.drawWidget();
    }
    return *this;
}

/** Change item's enabled status. */
WOptionGrid::Ref&
WOptionGrid::Ref::enable(bool flag)
{
    if (pItem && pItem->enabled != flag) {
        pItem->enabled = flag;
        pItem->button.setState(UIWidget::st_Disabled, !flag);
        pItem->parent.drawWidget();
    }
    return *this;
}


/** Constructor.
    \param leftWidth Width of left (label) column, in pixels
    \param rightWidth Width of right (value) column, in pixels */
WOptionGrid::WOptionGrid(int leftWidth, int rightWidth)
    : leftWidth(leftWidth),
      rightWidth(rightWidth),
      items()
{ }

WOptionGrid::~WOptionGrid()
{ }

/** Add an item.
    \param id Identifier to find the item later
    \param key Key (an upper-case letter)
    \param label Label (a string not ending in ":")
    \return reference to new item */
WOptionGrid::Ref
WOptionGrid::addItem(int id, char key, string_t label)
{
    Item* pItem = new Item(*this, id, key, label);
    items.push_back(pItem);

    const int ly = font_heights[FONT_TITLE];
    add(pItem->button, GfxRect(getExtent().x, getExtent().y + (ly*(items.size()-1)) + 1, ly-2, ly-2));
    return pItem;
}

/** Find an item, given an Id.
    The reference allows modifying the item.
    If the item was not found, the reference will "swallow" calls.
    \return reference */
WOptionGrid::Ref
WOptionGrid::find(int id)
{
    for (size_t i = 0, n = items.size(); i < n; ++i) {
        if (items[i]->id == id) {
            return items[i];
        }
    }
    return Ref();
}

void
WOptionGrid::drawContent(GfxCanvas& can)
{
    const GfxRect ext = getExtent();
    const int ly = font_heights[FONT_TITLE];

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    for (size_t i = 0; i < items.size(); ++i) {
        // Fetch item
        const Item& it = *items[i];
        int y = ext.y + (ly * (2*i+1)) / 2;

        // Left side
        ctx.useFont(*fonts[FONT_NORMAL]);
        ctx.setTextAlign(0, 1);
        ctx.setColor(it.enabled ? UIColor::tc_Static : UIColor::tc_Faded);
        ctx.setTransparentBackground();
        outText(ctx, ext.x + ly + 5, y, it.label + ":");

        // Right side
        ctx.useFont(*fonts[it.font]);
        ctx.setTextAlign(2, 1);
        ctx.setColor(it.enabled ? UIColor::tc_Green : UIColor::tc_Faded);
        ctx.setSolidBackground();
        outTextF(ctx, ext.x + ext.w, y, rightWidth, it.value);
    }

    can.update(ext);
}

void
WOptionGrid::getLayoutInfo(LayoutInfo& info)
{
    const int ly = font_heights[FONT_TITLE];
    const int width = ly       /* buttons */
        + 5                    /* gap */
        + leftWidth            /* left text */
        + 10                   /* estimation for ":" */
        + 10                   /* gap */
        + rightWidth;
    const int height = ly * items.size();

    info.pref_size = info.min_size = GfxPoint(width, height);
    info.flags = li_FlexibleH;
}

bool
WOptionGrid::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        for (size_t i = 0; i < items.size(); ++i) {
            if (event.key.code == uint32_t(items[i]->key) && items[i]->enabled) {
                sig_click.raise(items[i]->id);
                return true;
            }
        }
    }
    return UIWidget::handleEvent(event, second_pass);
}
