/**
  *  \file game/bbox.cc
  *  \brief Starchart Bounding Box
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "game/bbox.h"
#include "game/univ.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/ion.h"
#include "game/mine.h"
#include "game/ufo.h"
#include "game/drawing.h"

/******************************* GChartBBox ******************************/

/*! \class GChartBBox
    \brief Computing a Starchart Bounding Box

    This class provides functions to compute a bounding box. The sequence of
    events is:
    - add points using addPoint(), addCircle()
    - call initMaxZoom(), initMinZoom() to make the final computations */

/** Initialize bounding box from game universe. */
void
GChartBBox::initFromUniverse(const GUniverse& univ)
{
    min = chart_conf.getMinimumCoordinates();
    max = chart_conf.getMaximumCoordinates();

    /* Planets */
    {
        const GAnyPlanetType& pt = univ.ty_any_planets;
        for (int i = pt.findNextIndex(0); i != 0; i = pt.findNextIndex(i)) {
            addPoint(univ.getPlanet(i).getPos());
        }
    }

    /* Ships */
    {
        const GAnyShipType& st = univ.ty_any_ships;
        for (int i = st.findNextIndex(0); i != 0; i = st.findNextIndex(i)) {
            addPoint(univ.getShip(i).getPos());
        }
    }

    /* Storms */
    {
        const GIonStormType& it = univ.ty_ion_storms;
        for (int i = it.findNextIndex(0); i != 0; i = it.findNextIndex(i)) {
            const GIonStorm& st = univ.getIonStorm(i);
            addCircle(st.getPos(), st.getRadius());
        }
    }

    /* Minefields */
    for (GMinefieldType::const_iterator i = univ.ty_minefields.begin(); i != univ.ty_minefields.end(); ++i) {
        addCircle(i->getPos(), i->getRadius());
    }

    /* Drawings */
    const GDrawingContainer& draw = univ.getDrawings();
    for (GDrawingContainer::const_iterator i = draw.begin(); i != draw.end(); ++i) {
        addDrawing(**i);
    }

    /* Ufos */
    const GUfoType& ufos = univ.ty_ufos;
    for (GUfoType::const_iterator i = ufos.begin(); i != ufos.end(); ++i) {
        addCircle(i->getPos(), i->getRadius());
        if (GUfo* other_end = i->getOtherEnd()) {
            addCircle(other_end->getPos().getSimpleNearestAlias(i->getPos()), other_end->getRadius());
        }
    }

    initMinZoom();
    initMaxZoom();
}

/** Compute minimum zoom level. The minimum zoom is a soft limit;
    below the limit, the scrolling map is replaced by a fixed one. */
void
GChartBBox::initMinZoom()
{
    /* Minimum zoom. At zoom smaller 1:N, switch to overview chart */
    min_zoom = 2;
    // FIXME: dummy
    // CheckMinZoom(SCREEN_WIDTH DIV 2, ChartBX1-ChartConf.CenterX);
    // CheckMinZoom(SCREEN_WIDTH DIV 2, ChartConf.CenterX-ChartBX0);
    // CheckMinZoom((SCREEN_HEIGHT-80) DIV 2, ChartBY1-ChartConf.CenterY);
    // CheckMinZoom((SCREEN_HEIGHT-80) DIV 2, ChartConf.CenterY-ChartBY0);
    if (min_zoom > MAX_ZOOM)
        min_zoom = MAX_ZOOM;
}

/** Compute maximum zoom level. The maximum zoom level is a hard
    limit; under no conditions the hard limit may be exceeded to avoid
    integer overflow. */
void
GChartBBox::initMaxZoom()
{
    max_zoom = INT_MAX / std::max(max.x, max.y);
    if (max_zoom > MAX_ZOOM)
        max_zoom = MAX_ZOOM;
}

/** Include a point in the bounding box. */
void
GChartBBox::addPoint(GPoint pt)
{
    if (min.x > pt.x)
        min.x = pt.x;
    if (min.y > pt.y)
        min.y = pt.y;
    if (max.x < pt.x)
        max.x = pt.x;
    if (max.y < pt.y)
        max.y = pt.y;
}

/** Include a circle in the bounding box. */
void
GChartBBox::addCircle(GPoint pt, int radius)
{
    addPoint(GPoint(pt.x - radius, pt.y - radius));
    addPoint(GPoint(pt.x + radius, pt.y + radius));
}

/** Include a drawing in the bounding box. */
void
GChartBBox::addDrawing(const GDrawing& d)
{
    switch (d.getType()) {
     case LineDrawing:
     case RectangleDrawing:
        /* line/rectangle: include endpoints */
        addPoint(d.getPos());
        addPoint(d.getPos2());
        break;
     case CircleDrawing:
        /* circle */
        addCircle(d.getPos(), d.getCircleRadius());
        break;
     case MarkerDrawing:
        /* marker. We assume that all markers have a radius of 10 or less.
           This does not consider the optional text of the marker, though. */
        addCircle(d.getPos(), 10);
        break;
    }
}
