/**
  *  \file int/context.h
  *  \brief Interpreter: Contexts
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_CONTEXT_H
#define PCC_V2_INT_CONTEXT_H

#include "cpluslib/types.h"
#include "cpluslib/object.h"
#include "cpluslib/string.h"
#include "int/value.h"

struct IntExecutionFrame;
class IntPropertyAcceptor;
class GObject;
class IntNameQuery;

/** Context for name lookup. A context provides a means for looking up and
    dealing with local names, and possibly iteration through objects. */
class IntContext : public IntValue {
 public:
    /** Look up a symbol by its name.
        \return index for symbol (>= 0), or -1 for failure. */
    virtual int32_t lookup(const IntNameQuery& name) = 0;

    /** Set value by its index. The parameter is owned by the caller. */
    virtual void set(int32_t index, IntValue* value) = 0;

    /** Get value by its index. The returned value must be newly allocated,
        caller assumes responsibility. */
    virtual IntValue* get(int32_t index) = 0;

    /** Advance to next object. Return true on success, false on failure. */
    virtual bool next() = 0;

    /** Clone this context. */
    virtual IntContext* clone() = 0;

    /** Get associated game object. This is used for information purposes,
        and for type switches in various GUI function bindings. This may return
        null if this context is not associated with a game object. */
    virtual GObject* getObject() = 0;

    /** Enumerate properties. Call acceptor.addProperty for every property.
        \param acceptor Acceptor object */
    virtual void enumProperties(IntPropertyAcceptor& acceptor) = 0;

    // IntValue:
    // string_t toString(bool readable);
    // void store(IntTagNode& sv, Stream& aux);
};

/** Single context. This is a context that tells its user that it does not
    support iteration. */
class IntSingleContext : public IntContext {
 public:
    bool next();
};

/** Local variables for an execution frame. Provides access to the local variables
    of an executing stack frame. */
class IntExecutionFrameContext : public IntSingleContext {
    IntExecutionFrame& frame;
 public:
    IntExecutionFrameContext(IntExecutionFrame& frame);
    ~IntExecutionFrameContext();

    int32_t   lookup(const IntNameQuery& name);
    void      set(int32_t index, IntValue* name);
    IntValue* get(int32_t index);
    IntExecutionFrameContext* clone();
    GObject*  getObject();

    string_t  toString(bool readable);
    void      store(IntTagNode& sv, Stream& aux);
    void      enumProperties(IntPropertyAcceptor& acceptor);
};

#endif
