/**
  *  \file int/contextio.cc
  *  \brief Interpreter: Context I/O
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/contextio.h"
#include "int/context.h"
#include "int/if/shipif.h"
#include "int/if/planetif.h"
#include "int/if/mineif.h"
#include "int/if/ionif.h"
#include "int/if/hullif.h"
#include "int/if/specif.h"
#include "game/spec.h"
#include "io/memstream.h"
#include "io/io.h"
#include "util/nls.h"
#include "util/except.h"
#include "int/if/globalif.h"
#include "int/vmio.h"
#include "int/mutex.h"
#include "int/if/playerif.h"
#include "int/if/iterif.h"

/** Load context value from stream.
    \param sv Tag node read from header
    \param aux Auxiliary data can be read here */
IntContext*
loadContext(const IntTagNode& sv, Stream& aux)
{
    switch (sv.tag) {
     case Tag_Ship:
        // Ship. Check range.
        if (sv.value > 0 && sv.value <= NUM_SHIPS)
            return new IntShipContext(sv.value);
        else
            return 0;
     case Tag_Planet:
        // Planet. Check range.
        if (sv.value > 0 && sv.value <= NUM_PLANETS)
            return new IntPlanetContext(sv.value);
        else
            return 0;
     case Tag_Minefield:
        // We currently have no constraints on minefield Ids
        return new IntMinefieldContext(sv.value);
     case Tag_Ion:
        // Ion storm. Check range.
        if (sv.value > 0 && sv.value <= NUM_STORMS)
            return new IntIonContext(sv.value);
        else
            return 0;
     case Tag_Hull:
        // Hull. Check range.
        if (sv.value > 0 && int(sv.value) <= getNumHulls())
            return new IntHullContext(sv.value);
        else
            return 0;
     case Tag_Engine:
        // Engine. Check range.
        if (sv.value > 0 && sv.value <= NUM_ENGINES)
            return new IntEngineContext(sv.value);
        else
            return 0;
     case Tag_Beam:
        // Beam. Check range.
        if (sv.value > 0 && sv.value <= NUM_BEAMS)
            return new IntBeamContext(sv.value);
        else
            return 0;
     case Tag_Torpedo:
        // Torpedo. Check range.
        if (sv.value > 0 && sv.value <= NUM_TORPS)
            return new IntTorpedoContext(sv.value);
        else
            return 0;
     case Tag_Launcher:
        // Torpedo launcher. Check range.
        if (sv.value > 0 && sv.value <= NUM_TORPS)
            return new IntLauncherContext(sv.value);
        else
            return 0;
     case Tag_Frame:
        // Frame. Depends on current process.
        if (IntVMLoadContext* vlc = IntVMLoadContext::getCurrentInstance())
            return vlc->makeFrameContext(sv.value);
        else
            return 0;
     case Tag_Global:
        // Global context; unique.
        return new IntGlobalContext();
     case Tag_Mutex:
        // Mutex context.
        return IntMutexContext::load(sv.value, aux);
     case Tag_Iterator:
        // Iterator.
        if (GObjectType* typ = getObjectTypeFromIteratorId(sv.value)) {
            return new IntIteratorContext(*typ, sv.value);
        } else {
            return 0;
        }
     case Tag_Player:
        // Player. Check range.
        if (sv.value > 0 && sv.value <= NUM_PLAYERS) {
            return new IntPlayerContext(sv.value);
        } else {
            return 0;
        }
     default:
        return 0;
    }
}

/** Save context list into stream.
    \param out      [out] Stream
    \param contexts [in] Contexts to save
    \throws FileProblemException upon I/O error
    \throws IntError if a context cannot be serialized

    This is a very stripped-down version of IntDataSegment::save,
    which assumes contexts are never null. */
void
saveContexts(Stream& out, const ptr_vector<IntContext>& contexts)
{
    // Collect headers in one stream, aux data in another
    MemoryStream headers, aux;

    for (uint32_t i = 0; i < contexts.size(); ++i) {
        IntTagNode sv;
        contexts[i]->store(sv, aux);
        storeStructureT(headers, sv);
    }

    // Generate output
    out.writeT(headers.getBuffer().ptr(), headers.getBuffer().getSize());
    out.writeT(aux.getBuffer().ptr(),     aux.getBuffer().getSize());
}

/** Load context list from stream.
    \param out      [in] Stream to read from
    \param slots    [in] Number of contexts to load
    \param contexts [out] Contexts will be added here
    \throws FileFormatException if an invalid context is encountered

    This is a very stripped-down version of IntDataSegment::load,
    which assumes to find only contexts, no scalars or other values. */
void
loadContexts(Stream& in, uint32_t slots, ptr_vector<IntContext>& contexts)
{
    uint32 headerSize = slots * IntTagNode::size;
    Array<char> headers(headerSize);

    // Read headers
    in.readT(headers.ptr(), headerSize);

    // Read elements
    for (uint32_t i = 0; i < slots; ++i) {
        // Read tag node
        IntTagNode sv;
        getStructure(headers.ptr() + (IntTagNode::size * i), sv);

        IntContext* cv = loadContext(sv, in);
        if (!cv)
            throw FileFormatException(in, _("Invalid value in context list; file probably written by newer version of PCC"));

        contexts.push_back(cv);
    }
}
