/**
  *  \file io/streams.h
  *  \brief Stream Implementation
  *
  *  The stream base class.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_FILEIO_STREAMS_H
#define PCC_V2_FILEIO_STREAMS_H

#include <cstddef>
#include "cpluslib/smartptr.h"
#include "cpluslib/string.h"
#include "cpluslib/buffer.h"
#include "cpluslib/types.h"

class ChildStream;
class ProgressMonitor;

class Stream {
    friend class ChildStream;
    ChildStream* first_child;
    ChildStream* active_child;

 protected:
    std::size_t notReadable();
    std::size_t notWritable();
    void notSeekable();
    void invalidFilePosition();

 public:
    enum {
        C_READ   = 1,           ///< Stream can be read
        C_WRITE  = 2,           ///< Stream can be written to
        C_SEEK   = 4,           ///< Stream is seekable
        C_CREATE = 8            ///< (open) Create file
    };

    /** File position/size type. Must be compatible with signed long.
        "Compatible" means it must support most operations supported by
        a long, but it need not necessarily fit into a long.
        This allows it to be replaced by a class (for large-file support),
        but still keep stuff in source like `seek(0)' or `if(pos < 0)'. */
    typedef long fpos_t;

    Stream() throw();
    virtual ~Stream();

    virtual std::size_t read(char* buffer, std::size_t max) = 0;
    virtual std::size_t write(const char* buffer, std::size_t max) = 0;
    virtual int getCapabilities() const = 0;
    virtual void seek(fpos_t pos) = 0;
    virtual fpos_t getPos() = 0;
    virtual fpos_t getSize() = 0;

    virtual string_t getName() = 0;

    virtual Ptr<Stream> dup();

    bool canDo(int what);

    void readBuffer(Buffer<char>& buffer);

    void readT(char* buffer, std::size_t count);
    void writeT(const char* buffer, std::size_t count);

    void copyFrom(Stream& other, ProgressMonitor& pm);
    void copyFrom(Stream& other, uint32_t size, ProgressMonitor& pm);

    /* File Factory. The platform-specific code must override
       the openFile(string_t,int) version. The other one is provided
       for convenience. */
    static Ptr<Stream> openFile(const string_t& name, int mode);
    static Ptr<Stream> openFile(const char* name, int mode);

    /* convenience functions, similar to Directory::openFileNT */
    static Ptr<Stream> openFileNT(const string_t& name, int mode);
};

//! "Window" into a stream.
class SubStreamBase : public Stream {
    fpos_t origin, length, position;
 protected:
    Stream& worker;
    int op_mask;

    void setRange(fpos_t origin, fpos_t length);
    void setDisallowedOperations(int aop) { op_mask = aop; }
    SubStreamBase(Stream& worker);
 public:
    std::size_t read(char* buffer, std::size_t max);
    std::size_t write(const char* buffer, std::size_t max);
    int getCapabilities() const;
    void seek(fpos_t pos);
    fpos_t getPos();
    fpos_t getSize();
    string_t getName();
};

//! "Window" into a stream -- user-callable.
class SubStream : public SubStreamBase {
 public:
    SubStream(Stream& worker);
    // gcc doesn't eat using-declarations :(
    // using setRange;
    // using setDisallowedOperations;
    void setRange(fpos_t org, fpos_t len) { SubStreamBase::setRange(org, len); }
    void setDisallowedOperations(int aop) { SubStreamBase::setDisallowedOperations(aop); }
};

class ChildStream : public Stream {
    friend class Stream;
    Stream* parent;
    ChildStream* next_child;
    ChildStream** prev_child;

    fpos_t current_pos;

    bool isActive() const throw();
    bool activate();
    void deactivate();
    void disconnect() throw();
 public:
    ChildStream(Stream& parent);
    ChildStream() throw();
    ~ChildStream();

    std::size_t read(char* buffer, std::size_t max);
    std::size_t write(const char* buffer, std::size_t max);
    int getCapabilities() const;
    void seek(fpos_t pos);
    fpos_t getPos();
    fpos_t getSize();

    string_t getName();

    Ptr<Stream> dup();
};

#endif
