#!/usr/bin/perl -w
#
#  Quick & Dirty Object File / VM Dump
#
#  Usage:
#    perl c2vmdump.pl [FILE.qc|scriptX.cc...]
#

use strict;

# Object types
my %objectTypes = (
                   1 => "Bytecode Object (BCO)",
                   2 => "Process",
                   3 => "Frame",
                   4 => "Array",
                   5 => "Hash",
                   6 => "Structure",
                   7 => "Type"
                  );

# Property parsers. Indexed by object type and property number.
my %propertyParsers = (
                       1 => {
                             1 => { name=>"Header",       func=>\&parseBCOHeader },
                             2 => { name=>"Literal Pool", func=>\&parseDataSegment, arg=>'lit' },
                             3 => { name=>"Name Pool",    func=>\&parseNameList,    arg=>'sym' },
                             4 => { name=>"Code",         func=>\&parseCode },
                             5 => { name=>"Local Names",  func=>\&parseNameList },
                             6 => { name=>"Name Hint",    func=>\&parseString,      saveAs=>'bco' },
                             7 => { name=>"File Name",    func=>\&parseString },
                             8 => { name=>"Line Numbers", func=>\&parseLineNumbers }
                            },
                       2 => {
                             1 => { name=>"Header",       func=>\&parseProcessHeader },
                             2 => { name=>"Name",         func=>\&parseString },
                             3 => { name=>"Frames",       func=>\&parseEmbeddedObjects },
                             4 => { name=>"Contexts",     func=>\&parseDataSegment },
                             5 => { name=>"Exceptions",   func=>\&parseExceptions },
                             6 => { name=>"Values",       func=>\&parseDataSegment },
                            },
                       3 => {
                             1 => { name=>"Header",       func=>\&parseFrameHeader },
                             2 => { name=>"Values",       func=>\&parseDataSegment },
                             3 => { name=>"Names",        func=>\&parseNameList },
                            },
                       4 => {
                             1 => { name=>"Dimensions",   func=>\&parseIntList },
                             2 => { name=>"Values",       func=>\&parseDataSegment },
                            },
                       5 => {
                             1 => { name=>"Keys",         func=>\&parseNameList },
                             2 => { name=>"Values",       func=>\&parseDataSegment },
                            },
                       6 => {
                             1 => { name=>"Header",       func=>\&parseStructureHeader },
                             2 => { name=>"Values",       func=>\&parseDataSegment },
                            },
                       7 => {
                             1 => { name=>"Member Names", func=>\&parseNameList },
                            },
                      );

# Data types, for data segment parsing. Those are the tags stored in
# the second byte of each slot.
my %dataTypes = (
                 0 => "0.0",
                 1 => "NULL",
                 2 => "Int(%d)",
                 3 => "Bool(%d)",
                 # 4 => short string
                 5 => "Float32",
                 # 6 => long string
                 7 => "BCO(%d)",
                 8 => "Array(%d)",
                 # 9 => blob, like long string
                 10 => "File(#%d)",
                 11 => "Hash(%d)",
                 12 => "Struct(%d)",
                 13 => "Type(%d)",
                 0x80 => "Ship(%d)",
                 0x81 => "Planet(%d)",
                 0x82 => "Minefield(%d)",
                 0x83 => "Ion(%d)",
                 0x84 => "Hull(%d)",
                 0x85 => "Engine(%d)",
                 0x86 => "Beam(%d)",
                 0x87 => "Torpedo(%d)",
                 0x88 => "Launcher(%d)",
                 0x89 => "Frame(%d)",
                 0x8A => "Global",
                 # 8B => lock
                );

my %processKinds = (
                    0 => 'default',
                    1 => 'ship task',
                    2 => 'planet task',
                    3 => 'base task'
                   );

my %instructionTable;
my %decoderTable;
my %collectedNameHints;

initInstructionTable();
initDecoderTable();

foreach (@ARGV) {
    if (open FILE, "< $_") {
        %collectedNameHints = ();
        print "$_:\n";
        binmode FILE;
        dumpFile();
        close FILE;
    } else {
        print STDERR "$_: $!\n";
    }
}

sub dumpFile {
    # Read header
    if (!dumpHeader()) {
        print "  File does not have valid/supported header; file not interpreted.\n";
        return;
    }

    # Read content
    my $content = join "", <FILE>;
    parseObjects($content, "  ");
}

sub dumpHeader {
    my $head;
    read(FILE, $head, 4) == 4                or return 0;
    if ($head eq 'CCvm') {
        read(FILE, $head, 26, 4) == 26       or return 0;
        my ($sig, $time, $sig2, $version, $bytes, $turn, $player) = unpack "A4A18CCvvv", $head;
        if ($bytes >= 4 && $version >= 100) {
            print <<EOF;
  Header: VM file
    Signature:    '$sig'
    Timestamp:    '$time'
    Version:      $version
    Header Size:  $bytes
    Turn Number:  $turn
    Player:       $player
EOF
            return 1;
        } else {
            return 0;
        }
    } elsif ($head eq 'CCob') {
        read(FILE, $head, 10, 4) == 10       or return 0;
        my ($sig, $u1, $version, $u2, $bytes, $initBCO) = unpack "A5CCCvV", $head;
        if ($sig eq 'CCobj' && $version >= 100 && $bytes >= 4) {
            print <<EOF;
  Header: Object File
    Signature:    '$sig'
    Version:      $version
    Header Size:  $bytes
    Entry Point:  BCO $initBCO
EOF
            return 1;
        } else {
            return 0;
        }
    } else {
        return 0;
    }
}

sub parseObjects {
    my $content = shift;
    my $indent = shift;
    my $contentPos = 0;
    while ($contentPos < length($content)) {
        # Parse header
        my $avail = length($content) - $contentPos;
        if ($avail < 16) {
            print "ERROR: partial object ($avail bytes, expecting header)\n";
            return;
        }
        my ($type, $id, $size, $nprop) = unpack "VVVV", substr($content, $contentPos, 16);
        $contentPos += 16;
        $avail -= 16;

        # Validate header
        if ($avail < $size || $size <= 8*$nprop) {
            print "ERROR: partial object ($avail bytes, expecting property list)\n";
            return;
        }

        print $indent, "Object: ", formatEnum($type, \%objectTypes), ", Id $id, $nprop properties, $size bytes\n";

        # Read properties
        my @propCount;
        my @propSize;
        my $totalPropSize = 0;
        foreach (1 .. $nprop) {
            push @propCount, unpack "V", substr($content, $contentPos, 4);
            $contentPos += 4;
            $size -= 4;
            push @propSize, unpack "V", substr($content, $contentPos, 4);
            $totalPropSize += $propSize[-1];
            $contentPos += 4;
            $size -= 4;
        }

        # Validate
        if ($totalPropSize > $size) {
            print "ERROR: invalid object: total property size $totalPropSize exceeds room $size\n";
            return;
        }

        # Parse properties
        my $pRelay = {};
        foreach (0 .. $#propSize) {
            if ($propSize[$_] != 0) {
                if (!exists $propertyParsers{$type}{$_}) {
                    print "$indent  Property: Unknown (=$_), count $propCount[$_], size $propSize[$_]\n";
                } else {
                    print "$indent  Property: $propertyParsers{$type}{$_}{name} (=$_), count $propCount[$_], size $propSize[$_]\n";
                    $propertyParsers{$type}{$_}{func}("$indent    ",
                                                      substr($content, $contentPos, $propSize[$_]),
                                                      $propCount[$_],
                                                      $pRelay,
                                                      $propertyParsers{$type}{$_}{arg});
                    if (exists $propertyParsers{$type}{$_}{saveAs}) {
                        $collectedNameHints{$propertyParsers{$type}{$_}{saveAs}.$id} = substr($content, $contentPos, $propSize[$_]);
                    }
                }
                $contentPos += $propSize[$_];
            }
        }

        if ($totalPropSize < $size) {
            print "WARNING: ", $size - $totalPropSize, " unused bytes\n";
            $contentPos += $size - $totalPropSize;
        }
    }
}

#
#  Property Parsers
#

sub parseBCOHeader {
    my ($indent, $data, $count, $pRelay) = @_;
    if (length($data) < 8) {
        print "WARNING: object too small\n";
    } else {
        my ($flags, $min, $max, $labels) = unpack "v4", $data;
        printf "%sFlags:     %04X\n", $indent, $flags;
        printf "%sArgs:      %d..%d\n", $indent, $min, $max;
        printf "%sLabels:    %d\n", $indent, $labels;
    }
}

sub parseString {
    my ($indent, $data, $count, $pRelay) = @_;
    printf "%s'%s'\n", $indent, sanitize($data);
}

sub parseNameList {
    my ($indent, $data, $count, $pRelay, $arg) = @_;
    my $ptr = 0;
    my $i = 0;
    while ($i < $count) {
        if ($ptr > length($data)) {
            print "WARNING: object too small\n";
            return;
        }
        my $length = ord(substr($data, $ptr, 1)); ++$ptr;
        if ($ptr + $length > length($data)) {
            print "WARNING: object too small\n";
            return;
        }
        my $nr = $i++;
        my $str = sanitize(substr($data, $ptr, $length));
        printf "%s%d: '%s'\n", $indent, $nr, $str;
        if (defined($arg)) {
            $pRelay->{$arg}{$nr} = $str;
        }
        $ptr += $length;
    }
}

sub parseEmbeddedObjects {
    my ($indent, $data, $count, $pRelay) = @_;
    parseObjects($data, $indent);
}

sub parseDataSegment {
    my ($indent, $data, $count, $pRelay, $arg) = @_;
    if (length($data) < 6*$count) {
        print "WARNING: object too small\n";
        return;
    }
    my $auxptr = 6*$count;
    for (my $i = 0; $i < $count; ++$i) {
        my ($tag, $value) = unpack "vV", substr($data, 6*$i, 6);
        my $fmt;
        if ($tag == 0x0400) {
            # short string
            if ($value == 0) {
                $fmt = "''";
            } else {
                if ($auxptr >= length($data)) {
                    print "WARNING: object too small\n";
                    return;
                }
                my $length = substr($data, $auxptr++, 1);
                if ($auxptr + $length >= length($data)) {
                    print "WARNING: object too small\n";
                    return;
                }
                $fmt = "'" . sanitize(substr($data, $auxptr, $length)) . "'";
                $auxptr += $length;
            }
        } elsif ($tag == 0x0600) {
            # long string
            if ($value > length($data) - $auxptr) {
                print "WARNING: object too small\n";
                return;
            }
            $fmt = "'" . sanitize(substr($data, $auxptr, $value)) . "'";
            $auxptr += $value;
        } elsif ($tag == 0x0900) {
            # blob, like long string
            if ($value > length($data) - $auxptr) {
                print "WARNING: object too small\n";
                return;
            }
            $fmt = "Blob($value bytes)";
            $auxptr += $value;
        } elsif ($tag == 0x8B00) {
            # lock
            if (length($data) - $auxptr < 8) {
                print "WARNING: object too small\n";
                return;
            }
            my ($nameLength, $noteLength) = unpack("VV", substr($data, $auxptr, 8));
            $auxptr += 8;
            if ($nameLength < 0 || $noteLength < 0 || length($data) - $auxptr < $nameLength + $noteLength) {
                print "WARNING: object too small\n";
                return;
            }
            my $name = substr($data, $auxptr, $nameLength); $auxptr += $nameLength;
            my $note = substr($data, $auxptr, $noteLength); $auxptr += $noteLength;
            $fmt = "Lock('" . sanitize($name) . "', '" . sanitize($note) . "', $value)";
        } elsif (($tag & 0x00FF) == 0) {
            my $a = $tag >> 8;
            if (exists($dataTypes{$a})) {
                $fmt = sprintf($dataTypes{$a}, $value);
            } else {
                $fmt = "Unknown($a,$value)";
            }
        } else {
            $fmt = "Float48";
        }
        printf "%s%d: %s\n", $indent, $i, $fmt;
        if (defined($arg)) {
            $pRelay->{$arg}{$i} = $fmt;
        }
    }
    if ($auxptr != length($data)) {
        print "WARNING: object too big\n";
    }
}

sub parseProcessHeader {
    my ($indent, $data, $count, $pRelay) = @_;
    if (length($data) < 4) {
        print "WARNING: object too small\n";
    } else {
        my ($pri, $flags, $ctos) = unpack "CCv", $data;
        printf "%sPriority:    %d\n", $indent, $pri;
        printf "%sKind:        %s\n", $indent, formatEnum($flags, \%processKinds);
        printf "%sContext TOS: %d\n", $indent, $ctos;
    }
}

sub parseFrameHeader {
    my ($indent, $data, $count, $pRelay) = @_;
    if (length($data) < 20) {
        print "WARNING: object too small\n";
    } else {
        my ($bco, $pc, $csp, $xsp, $flags) = unpack "VVVVV", $data;
        printf "%sBCO:          %s\n", $indent, formatNameHint($bco, "bco$bco");
        printf "%sPC:           %d\n", $indent, $pc;
        printf "%sContext SP:   %d\n", $indent, $csp;
        printf "%sException SP: %d\n", $indent, $xsp;
        printf "%sFlags:        %08X\n", $indent, $flags;
    }
}

sub parseExceptions {
    my ($indent, $data, $count, $pRelay) = @_;
    if (length($data) < 16*$count) {
        print "WARNING: object too small\n";
    } else {
        for (my $i = 0; $i < $count; ++$i) {
            printf "%sException Frame #%d\n", $indent, $i+1;
            my ($fsp,$csp,$vsp,$pc) = unpack "VVVV", substr($data, 16*$i, 16);
            printf "%s  Frame SP:   %d\n", $indent, $fsp;
            printf "%s  Context SP: %d\n", $indent, $csp;
            printf "%s  Value SP:   %d\n", $indent, $vsp;
            printf "%s  PC:         %d\n", $indent, $pc;
        }
    }
}

sub parseLineNumbers {
    my ($indent, $data, $count, $pRelay) = @_;
    if (length($data) < 8*$count) {
        print "WARNING: object too small\n";
    } else {
        for (my $i = 0; $i < $count; ++$i) {
            my ($adr,$line) = unpack "VVVV", substr($data, 8*$i, 8);
            printf "%s  %6d => line %d\n", $indent, $adr, $line;
        }
    }
}

sub parseIntList {
    my ($indent, $data, $count, $pRelay) = @_;
    if (length($data) < 4*$count) {
        print "WARNING: object too small\n";
    } else {
        my $i = 0;
        foreach (unpack "V*", $data) {
            printf "%s  %6d => %d\n", $indent, $i++, $_;
        }
    }
}

sub parseCode {
    my ($indent, $data, $count, $pRelay) = @_;
    if (length($data) < 4*$count) {
        print "WARNING: object too small\n";
    } else {
        for (my $i = 0; $i < $count; ++$i) {
            my ($arg,$minor,$major) = unpack "vCC", substr($data, 4*$i, 4);
            my $fmt = '?';
            my $comment;
            if (exists $decoderTable{$major}{$minor}) {
                my $m = $decoderTable{$major}{$minor}[0];
                my $k = $decoderTable{$major}{$minor}[1];
                if ($k eq 'label') {
                    if ($minor & 0x80) {
                        $fmt = sprintf("%-8s sym-%d", $m, $arg);
                    } else {
                        $fmt = sprintf("%-8s #%d", $m, $arg);
                    }
                } elsif ($k eq 'none' || ($k =~ /^\?/ && $arg == 0)) {
                    $fmt = $m;
                } else {
                    $fmt = sprintf("%-8s %d", $m, $arg);
                }
                if (($k eq 'sym' || $k eq 'lit')  && exists($pRelay->{$k}{$arg})) {
                    $comment = $pRelay->{$k}{$arg};
                }
            }

            my $line = sprintf "%s %4d: %02X.%02X.%04X    %s", $indent, $i, $major, $minor, $arg, $fmt;
            if (defined($comment)) {
                printf "%-45s ; %s\n", $line, $comment;
            } else {
                printf "%s\n", $line;
            }
        }
    }
}

sub parseStructureHeader {
    my ($indent, $data, $count, $pRelay) = @_;
    if (length($data) < 4) {
        print "WARNING: object too small\n";
    } else {
        my $id = unpack "V", $data;
        printf "%sType:        Type(%d)\n", $indent, $id;
    }
}

#
#  Formatting utilities
#

sub formatEnum {
    my ($value, $pInfo) = @_;
    if (exists($pInfo->{$value})) {
        "$pInfo->{$value} (=$value)"
    } else {
        "Unknown (=$value)";
    }
}

sub sanitize {
    my $str = shift;
    $str =~ s/\\/\\\\/g;
    $str =~ s/([^ -~])/sprintf('\\x%02X', ord($1))/eg;
    $str;
}

#
#  Disassembler utilities
#
#  We dissassemble by inverting the same table used in the assembler
#  (c2asm).
#

sub initInstructionTable {
    %instructionTable = (
                         # 0 - push
                         # 1 - binary
                         # 2 - unary
                         # 3 - ternary
                         # 4 - jump
                         nop     => [4,0,'none'],
                         label   => [4,0,'label'],
                         jt      => [4,1,'label'],
                         jf      => [4,2,'label'],
                         jtf     => [4,3,'label'],
                         je      => [4,4,'label'],
                         jte     => [4,5,'label'],
                         jfe     => [4,6,'label'],
                         j       => [4,7,'label'],
                         jneverp => [4,8,'label'],
                         jtp     => [4,9,'label'],
                         jfp     => [4,10,'label'],
                         jtfp    => [4,11,'label'],
                         jep     => [4,12,'label'],
                         jtep    => [4,13,'label'],
                         jfep    => [4,14,'label'],
                         jp      => [4,15,'label'],
                         catch   => [4,16,'label'],
                         jdz     => [4,17,'label'],

                         # 5 - indirect
                         # 6 - stack
                         dup     => [6,0,'int'],
                         drop    => [6,1,'int'],
                         swap    => [6,2,'int'],

                         # 7 - pop
                         # 8 - store
                         # 9 - memref
                         # 10 - dim
                         # 11 - special
                         suncatch    => [11,0,'none'],
                         sreturn     => [11,1,'int'],
                         swith       => [11,2,'none'],
                         sendwith    => [11,3,'none'],
                         sfirstindex => [11,4,'none'],
                         snextindex  => [11,5,'none'],
                         sendindex   => [11,6,'none'],
                         sevals      => [11,7,'int'],
                         sevalx      => [11,8,'none'],
                         sdefsub     => [11,9,'sym'],
                         sdefshipp   => [11,10,'sym'],
                         sdefplanetp => [11,11,'sym'],
                         sload       => [11,12,'none'],
                         sprint      => [11,13,'none'],
                         saddhook    => [11,14,'?int'], # optional arg; might be used later, but isn't used now
                         srunhook    => [11,15,'?int'], # optional arg; might be used later, but isn't used now
                         sthrow      => [11,16,'none'],
                         sterminate  => [11,17,'none'],
                         ssuspend    => [11,18,'none'],
                         snewarray   => [11,19,'int'],
                         smakelist   => [11,20,'int'],
                         snewhash    => [11,21,'?int'],
                         sinstance   => [11,22,'none'],
                         sresizearray => [11,23,'int'],
                         sbind       => [11,24,'int'],
                        );

    # Major 2 - unary
    my @unaryList = qw( not bool neg pos sin
                        cos tan zap abs exp
                        log bitnot isempty isnum isstr
                        asc chr str sqrt trunc
                        round ltrim rtrim lrtrim length
                        val trace not2 atom atomstr
                        keycreate keylookup inc dec isproc
                        filenr isarray );
    foreach (0 .. $#unaryList) {
        $instructionTable{'u'.lc($unaryList[$_])} = [2,$_,'none'];
    }

    # Major 1 - binary
    my @binaryList = qw( and or xor add sub
                         mul div idiv rem pow
                         concat concempty cmpeq cmpeqNC cmpne
                         cmpneNC cmple cmpleNC cmplt cmpltNC
                         cmpge cmpgeNC cmpgt cmpgtNC min
                         minNC max maxNC firststr firststrNC
                         reststr reststrNC findstr findstrNC bitand
                         bitor bitxor str atan lcut
                         rcut endcut strmult keyaddparent keyfind
                         arraydim );
    foreach (0 .. $#binaryList) {
        $instructionTable{'b'.lc($binaryList[$_])} = [1,$_,'none'];
    }

    # Major 3 - ternary
    my @ternaryList = qw( keyadd );
    foreach (0 .. $#ternaryList) {
        $instructionTable{'t'.lc($ternaryList[$_])} = [3,$_,'none'];
    }

    # Major 0,7,8,10 - "scope" operations
    foreach (['var',0,'sym'],  ['loc',1,'loc'],  ['top',2,'int'],
             ['glob',3,'int'], ['gvar',4,'sym'], ['lit',5,'lit'],
             ['int',6,'int'],  ['bool',7,'int'])
      {
          $instructionTable{'push'.$_->[0]}  = [0, $_->[1], $_->[2]];
          $instructionTable{'pop'.$_->[0]}   = [7, $_->[1], $_->[2]];
          $instructionTable{'store'.$_->[0]} = [8, $_->[1], $_->[2]];
          $instructionTable{'dim'.$_->[0]}   = [10, $_->[1], 'sym'];
      }

    # Major 5,9 - "i/m" operations
    foreach (["call"  => 0], ["load"  => 1], ["store"  => 2], ["pop"   => 3],
             ["proc"  => 4], ["pload" => 5], ["pstore" => 6], ["ppop"  => 7],
             ["fcall" => 8], ["func"  => 9], ["fstore" => 10], ["fpop" => 11])
      {
          $instructionTable{$_->[0]."ind"} = [5,$_->[1], 'int'];
          $instructionTable{$_->[0]."mem"} = [9,$_->[1], 'sym'];
      }
}

sub initDecoderTable {
    foreach (keys %instructionTable) {
        next if $_ eq 'nop';
        my $mnemo = $_;
        my $major = $instructionTable{$_}[0];
        my $minor = $instructionTable{$_}[1];
        my $arg   = $instructionTable{$_}[2];
        $decoderTable{$major}{$minor} = [$mnemo, $arg];
        if ($major == 4) {
            # Major 4 - jumps, those appear in two flavours
            $decoderTable{$major}{$minor | 0x80} = [$mnemo, $arg];
        }
    }
}

sub formatNameHint {
    my $int = shift;
    my $key = shift;
    if (exists $collectedNameHints{$key}) {
        sprintf "%-6d ; %s", $int, $collectedNameHints{$key};
    } else {
        $int;
    }
}
