/**
  *  \file u/t_game_ufo.cc
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "u/t_game.h"
#include "game/config.h"
#include "game/ufo.h"
#include "game/utildata.h"
#include "game/struct.h"

static const TDbUfo test_db_ufo = {
    60,                         // Id
    {
        1,                      // color
        "Wormhole #9",          // name
        "",                     // info1
        "",                     // info2
        2000, 2000,             // x,y
        0,                      // warp,
        -1,                     // heading
        150, 150,               // planet/ship_range
        5,                      // radius
        1,                      // type_code
    },
    9,                          // real_id
    10,                         // turn_last_seen,
    2000, 2000,                 // x/y_last_seen
    7, 7,                       // x/y_move
};

static const TUtil14Wormhole test_wh = {
    2000, 2000,                 // x,y
    5000,                       // mass
    2,                          // stable_code
    9,                          // worm_id
    60,                         // ufo_id
    1,                          // bidir
};

void
GameUfoTestSuite::doSingleMovementTest(int scan_x,
                                       int expected_result,
                                       int turns,
                                       const char* name)
{
    // Create Ufo from history database
    GUfo ufo(test_db_ufo.id);
    ufo.addHistoryData(test_db_ufo);

    // Add (modified) scan
    TUtil14Wormhole wh = test_wh;
    wh.x = scan_x;
    ufo.addWormholeData(wh, false, test_db_ufo.turn_last_seen + turns);

    // Perform postprocessing
    ufo.postprocess(test_db_ufo.turn_last_seen + turns);

    // Now check result
    TSM_ASSERT(name, ufo.isSeenThisTurn());
    TSM_ASSERT_EQUALS(name, ufo.getPos().x, scan_x);
    TSM_ASSERT_EQUALS(name, ufo.getMovementVector().x, expected_result);
}

#define XSTR(X) #X
#define STR(X) XSTR(X)
#define NAME __FILE__ ":" STR(__LINE__)

void
GameUfoTestSuite::testMovementGuessing()
{
    // Non-overlapping WrmDisplacement / WrmRandDisplacement
    config.WrmDisplacement.set(10);
    config.WrmRandDisplacement.set(2);

    // -----     -----     -----
    //   |---------|---------|
    doSingleMovementTest(1988, -10, 1, NAME);
    doSingleMovementTest(1989, -10, 1, NAME);
    doSingleMovementTest(1990, -10, 1, NAME);
    doSingleMovementTest(1991, -10, 1, NAME);
    doSingleMovementTest(1992, -10, 1, NAME);
    doSingleMovementTest(1993, -10, 1, NAME);
    doSingleMovementTest(1994, -10, 1, NAME);
    doSingleMovementTest(1995, -10, 1, NAME);
    doSingleMovementTest(1996, -10, 1, NAME);
    doSingleMovementTest(1997, -10, 1, NAME);
    doSingleMovementTest(1998,   0, 1, NAME);
    doSingleMovementTest(1999,   0, 1, NAME);
    doSingleMovementTest(2000,   0, 1, NAME);
    doSingleMovementTest(2001,   0, 1, NAME);
    doSingleMovementTest(2002,   0, 1, NAME);
    doSingleMovementTest(2003,  10, 1, NAME);
    doSingleMovementTest(2004,  10, 1, NAME);
    doSingleMovementTest(2005,  10, 1, NAME);
    doSingleMovementTest(2006,  10, 1, NAME);
    doSingleMovementTest(2007,  10, 1, NAME);
    doSingleMovementTest(2008,  10, 1, NAME);
    doSingleMovementTest(2009,  10, 1, NAME);
    doSingleMovementTest(2010,  10, 1, NAME);
    doSingleMovementTest(2011,  10, 1, NAME);
    doSingleMovementTest(2012,  10, 1, NAME);

    // Overlapping WrmDisplacement / WrmRandDisplacement
    config.WrmDisplacement.set(10);
    config.WrmRandDisplacement.set(7);

    //           ---------------
    // ---------------     ---------------
    //        |---------|---------|
    doSingleMovementTest(1983, -10, 1, NAME);
    doSingleMovementTest(1984, -10, 1, NAME);
    doSingleMovementTest(1985, -10, 1, NAME);
    doSingleMovementTest(1986, -10, 1, NAME);
    doSingleMovementTest(1987, -10, 1, NAME);
    doSingleMovementTest(1988, -10, 1, NAME);
    doSingleMovementTest(1989, -10, 1, NAME);
    doSingleMovementTest(1990, -10, 1, NAME);
    doSingleMovementTest(1991, -10, 1, NAME);
    doSingleMovementTest(1992, -10, 1, NAME);
    doSingleMovementTest(1993,   7, 1, NAME); // ambiguous, could be -7 or -10+3
    doSingleMovementTest(1994,   7, 1, NAME); // ambiguous
    doSingleMovementTest(1995,   7, 1, NAME); // ambiguous
    doSingleMovementTest(1996,   7, 1, NAME); // ambiguous
    doSingleMovementTest(1997,   7, 1, NAME); // ambiguous, could be -3 or -10+7
    doSingleMovementTest(1998,   0, 1, NAME);
    doSingleMovementTest(1999,   0, 1, NAME);
    doSingleMovementTest(2000,   0, 1, NAME);
    doSingleMovementTest(2001,   0, 1, NAME);
    doSingleMovementTest(2002,   0, 1, NAME);
    doSingleMovementTest(2003,   7, 1, NAME); // ambiguous, could be +7 or +10-3
    doSingleMovementTest(2004,   7, 1, NAME); // ambiguous                      
    doSingleMovementTest(2005,   7, 1, NAME); // ambiguous                      
    doSingleMovementTest(2006,   7, 1, NAME); // ambiguous                      
    doSingleMovementTest(2007,   7, 1, NAME); // ambiguous, could be +3 or +10-7
    doSingleMovementTest(2008,  10, 1, NAME);
    doSingleMovementTest(2009,  10, 1, NAME);
    doSingleMovementTest(2010,  10, 1, NAME);
    doSingleMovementTest(2011,  10, 1, NAME);
    doSingleMovementTest(2012,  10, 1, NAME);
    doSingleMovementTest(2013,  10, 1, NAME);
    doSingleMovementTest(2014,  10, 1, NAME);
    doSingleMovementTest(2015,  10, 1, NAME);
    doSingleMovementTest(2016,  10, 1, NAME);
    doSingleMovementTest(2017,  10, 1, NAME);
}
