/**
  *  \file util/pluginstall.cc
  *  \brief Plugin Installer
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "util/pluginstall.h"
#include "util/plugin.h"
#include "io/streams.h"
#include "util/progress.h"
#include "util/pluginmgr.h"
#include "util/except.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "io/zip.h"

static void
copyFile(ProgressMonitor& pm, Directory& dstDir, Directory& srcDir, const string_t fname)
{
    Ptr<Stream> inFile = srcDir.openFile(fname, Stream::C_READ);
    Ptr<Stream> outFile = dstDir.openFile(fname, Stream::C_CREATE);
    outFile->copyFrom(*inFile, pm);
}

static bool
eraseFile(Directory& dir, const string_t& name)
{
    try {
        dir.erase(name);
        return true;
    }
    catch (FileProblemException& e) {
        console.write(LOG_WARN, format("%s: %s", e.getFileName(), e.what()));
        return false;
    }
    catch (std::exception& e) {
        console.write(LOG_WARN, format("%s: %s", Directory::makeFileName(dir.getName(), name), e.what()));
        return false;
    }
}

namespace {
    PluginInstaller::ScanResult scanDirectory(Directory& dir, string_t& pluginName)
    {
        PluginInstaller::ScanResult result = PluginInstaller::NoPlugin;
        Ptr<Enumeration<Ptr<DirectoryEntry> > > content = dir.getEntries();
        while (content->hasMoreElements()) {
            Ptr<DirectoryEntry> elem = content->getNextElement();
            const string_t baseName = elem->getBaseName();
            if (baseName.size() > 4 && strCaseCmp(baseName.substr(baseName.size() - 4), ".c2p") == 0) {
                if (result == PluginInstaller::NoPlugin) {
                    result = PluginInstaller::OnePlugin;
                    pluginName = baseName;
                } else {
                    result = PluginInstaller::MultiplePlugins;
                }
            }
        }
        return result;
    }
}


/**************************** PluginInstaller ****************************/

/** Constructor.
    \param mgr PluginManager
    \param rootDir Plugin root directory (points to PluginManager's root)
    \param pm Progress Monitor */
PluginInstaller::PluginInstaller(PluginManager& mgr, Directory& rootDir, ProgressMonitor& pm)
    : manager(mgr),
      rootDir(rootDir),
      progress(pm),
      srcDir(),
      srcFile(),
      apPlug()
{ }

PluginInstaller::~PluginInstaller()
{ }

/** Prepare installation.
    Checks whether the file name refers to a file that can be installed as a plugin.
    If so, makes a proto-plugin and returns it.
    The plugin remains owned by the PluginInstaller.
    The caller can examine it.
    It can be installed by calling doInstall().

    \param fileName File to install
    \return proto-plugin or null */
Plugin*
PluginInstaller::prepareInstall(string_t fileName)
{
    // Reset
    apPlug.reset();
    srcDir = 0;
    srcFile = 0;

    // Open plugin
    string_t baseName = Directory::getFileName(fileName);
    string_t dirName = Directory::getDirName(fileName);
    srcFile = Stream::openFile(fileName, Stream::C_READ);
    srcDir = Directory::openDir(dirName);

    // Determine file type
    if (baseName.size() > 4 && strCaseCmp(baseName.substr(baseName.size() - 4), ".c2p") == 0) {
        // Regular *.c2p file
        string_t pluginName = baseName.substr(0, baseName.size() - 4);
        apPlug.reset(new Plugin(strUCase(pluginName)));
        apPlug->initFromPluginFile(dirName, baseName, *srcFile);
        srcFile->seek(0);
    } else if (baseName.size() > 4 && strCaseCmp(baseName.substr(baseName.size() - 4), ".res") == 0) {
        // *.res file with synthetic definition
        string_t pluginName = baseName.substr(0, baseName.size() - 4);
        apPlug.reset(new Plugin(strUCase(pluginName)));
        apPlug->initFromResourceFile(dirName, baseName);
        srcFile = 0;
    } else if (baseName.size() > 2 && strCaseCmp(baseName.substr(baseName.size() - 2), ".q") == 0) {
        // *.q file with synthetic definition
        string_t pluginName = baseName.substr(0, baseName.size() - 2);
        apPlug.reset(new Plugin(strUCase(pluginName)));
        apPlug->initFromScriptFile(dirName, baseName, *srcFile);
        srcFile = 0;
    } else if (baseName.size() > 4
               && (strCaseCmp(baseName.substr(baseName.size() - 4), ".zip") == 0
                   || strCaseCmp(baseName.substr(baseName.size() - 4), ".c2z") == 0))
    {
        // *.zip/*.c2z: zipped plugin
        Ptr<Directory> zip = ZipFile::open(srcFile, 0);
        string_t pluginFile;
        if (scanDirectory(*zip, pluginFile) == PluginInstaller::OnePlugin) {
            // Read definition
            Ptr<Stream> file = zip->openFile(pluginFile, Stream::C_READ);
            string_t pluginName = pluginFile.substr(0, pluginFile.size() - 4);
            apPlug.reset(new Plugin(strUCase(pluginName)));
            apPlug->initFromPluginFile(fileName, pluginFile, *file);

            // Set output
            srcFile = zip->openFile(pluginFile, Stream::C_READ);
            srcDir = zip;
        }
    } else {
        // skip
    }

    return apPlug.get();
}

/** Install the prepared plugin.
    Throws exceptions on errors.
    \param dry true to simulate only */
void
PluginInstaller::doInstall(bool dry)
{
    // If there already is a plugin with that name, uninstall it first
    if (Plugin* originalPlugin = manager.getPluginById(apPlug->getId())) {
        doRemove(originalPlugin, dry);
    }

    if (!dry) {
        // Create directory
        Ptr<DirectoryEntry> dirEntry = rootDir.getEntryByName(strLCase(apPlug->getId()));
        try {
            dirEntry->createAsDirectory();
        }
        catch (...) { }
        Ptr<Directory> dir = dirEntry->openDir();

        // OK, do it
        const Plugin::ItemList& items = apPlug->getItems();
        for (size_t i = 0, n = items.size(); i != n; ++i) {
            const Plugin::Item& item = items[i];
            switch (item.type) {
             case Plugin::PlainFile:
             case Plugin::ScriptFile:
             case Plugin::ResourceFile:
                copyFile(progress, *dir, *srcDir, item.name);
                break;

             case Plugin::Command:
                break;
            }
        }

        // Create c2p file
        Ptr<Stream> c2pFile = rootDir.openFile(strLCase(apPlug->getId()) + ".c2p", Stream::C_CREATE);
        if (srcFile) {
            c2pFile->copyFrom(*srcFile, progress);
        } else {
            apPlug->savePluginFile(*c2pFile);
        }

        // Move the plugin definition
        // (This is needed in case the PluginManager is long-lived,
        // so someone actually needs this.)
        apPlug->setBaseDirectory(dir->getName());
    }

    // Reset
    manager.addNewPlugin(apPlug.release());
    srcFile = 0;
    srcDir  = 0;
}

/** Check for installation ambiguities.
    An ambiguity is when the user chose a file to auto-convert,
    but there is a *.c2p he should probably use instead.
    \param out [out] Name of alternative *.c2p file
    \retval NoPlugin no ambiguity
    \retval OnePlugin there is one *.c2p file that could be installed instead, its name given in %out.
    \retval MultiplePlugins there are multiple *.c2p files that could be installed instead */
PluginInstaller::ScanResult
PluginInstaller::checkInstallAmbiguity(string_t& out)
{
    if (srcFile == 0 && srcDir != 0) {
        // We do not have a source file.
        // this means we are auto-converting and could be subject to ambiguities.
        return scanDirectory(*srcDir, out);
    } else {
        return NoPlugin;
    }
}

/** Remove a plugin.
    Deletes all associated files.
    \param pPlug Plugin
    \param dry true to simulate only */
bool
PluginInstaller::doRemove(Plugin* pPlug, bool dry)
{
    // Remove the plugin from the manager
    std::auto_ptr<Plugin> apPlug(manager.extractPlugin(pPlug));

    // Early exit
    if (apPlug.get() == 0 || dry) {
        return true;
    }

    // Remove
    bool err = false;
    try {
        // Remove the plugin directory content
        Ptr<Directory> dir = Directory::openDir(apPlug->getBaseDirectory());
        const Plugin::ItemList& items = apPlug->getItems();
        for (size_t ii = 0, nn = items.size(); ii < nn; ++ii) {
            const Plugin::Item& item = items[ii];
            switch (item.type) {
             case Plugin::PlainFile:
             case Plugin::ScriptFile:
             case Plugin::ResourceFile:
                if (!eraseFile(*dir, item.name)) {
                    err = true;
                }
                break;

             case Plugin::Command:
                break;
            }
        }

        // Remove the plugin definition file
        const string_t& file = apPlug->getDefinitionFileName();
        if (!file.empty()) {
            if (!eraseFile(rootDir, file)) {
                err = true;
            }
        }

        // Remove the plugin directory
        string_t dirName = strLCase(apPlug->getId());
        if (!eraseFile(rootDir, dirName)) {
            err = true;
        }
    }
    catch (std::exception& e) {
        console.write(LOG_ERROR, e.what());
        err = true;
    }
    if (err) {
        console.write(LOG_WARN, format(_("Uninstallation of '%s' might be incomplete."), apPlug->getId()));
    }
    return !err;
}
